import sys
from importlib import util
from copy import deepcopy
from os import path, listdir

from .socket import SendSocket
from .module import Module
from .message import *
from .errors import ExtensionAlreadyLoaded, ExtensionNotFound, ExtensionFailed, NoEntryPointError 
from .log import Log


class Channel():

    name: str
    __modules: dict
    __modules: dict
    socket: SendSocket

    def __init__(self, name: str = "undefined"):
        self.bridge = None
        self.name = name
        self.__modules = {}
        self.__modules = {}

        self.log = Log(instance=name, client="Twitch Channel")
        pass

    # Public modules for use in the Bridge and in modules

    def propogate(self, bridge, socket: SendSocket, channel_name: str):
        self.name = channel_name
        self.log.set_instance(channel_name)
        newchannel = deepcopy(self)
        newchannel.__cache_modules()
        newchannel.socket = socket
        newchannel.bridge = bridge
        return newchannel
        


    def send(self, message):
        self.socket.send_message(self.name, message)

    def load_module(self, module):
        self.__modules[module.name] = module

    def on_new_connection(self, channel: str):
        self.log.debug("`on_new_connection` invoked.")
        for module in self.__get_module_with_handler("on_new_connection"):
            module.on_new_connection(channel)
        



    # Callbacks used by the bridge for all Messages sent via the inbound socket.

    ### to do: add the rest of the message types in here

    # Base Callback
    def on_raw(self, data: RawMessage):
        self.log.debug("`on_raw` invoked.")
        for module in self.__get_module_with_handler("on_raw"):
            module.on_raw(data)

    # IRC: Membership Callbacks

    def on_join(self, data: Join):
        self.log.debug("`on_join` invoked.")
        for module in self.__get_module_with_handler("on_join"):
            module.on_join(data)

    def on_part(self, data: Part):
        self.log.debug("`on_part` invoked.")
        for module in self.__get_module_with_handler("on_part"):
            module.on_part(data)

    def on_mode(self, data: Mode):
        self.log.debug("`on_mode` invoked.") 
        for module in self.__get_module_with_handler("on_mode"):
            module.on_mode(data)

    def on_names(self, data: Names):
        self.log.debug("`on_names` invoked.")
        for module in self.__get_module_with_handler("on_names"):
            module.on_names(data)

    # IRC: module & Tag Callbacks

    def on_clearchat(self, data: ClearChat):
        self.log.debug("`on_clearchat` invoked.")
        for module in self.__get_module_with_handler("on_clearchat"):
            module.on_clearchat(data)

    def on_clearmessage(self, data: ClearMessage):
        self.log.debug("`on_clearmessage` invoked.")
        for module in self.__get_module_with_handler("on_clearmessage"):
            module.on_clearmessage(data)
    
    def on_hosttarget(self, data: HostTarget):
        self.log.debug("`on_hosttarget` invoked.")
        for module in self.__get_module_with_handler("on_hosttarget"):
            module.on_hosttarget(data)

    def on_notice(self, data: Notice):
        self.log.debug(f"`on_notice` invoked. (not implemented)")
        for module in self.__get_module_with_handler("on_notice"):
            module.on_notice(data)

    def on_reconnect(self, data: Reconnect):
        self.log.debug("`on_reconnect` invoked.")
        for module in self.__get_module_with_handler("on_reconnect"):
            module.on_reconnect(data)

    def on_roomstate(self, data: RoomState):
        self.log.debug("`on_roomstate` invoked.")
        for module in self.__get_module_with_handler("on_roomstate"):
            module.on_roomstate(data)

    def on_userstate(self, data: UserState):
        self.log.debug("`on_userstate` invoked.")
        for module in self.__get_module_with_handler("on_userstate"):
            module.on_userstate(data)

    def on_globaluserstate(self, data: GlobalUserState):
        self.log.debug("`on_globaluserstate` invoked.")
        for module in self.__get_module_with_handler("on_globaluserstate"):
            module.on_globaluserstate(data)

    def on_privmessage(self, data: PrivateMessage):
        self.log.debug("`on_privmessage` invoked.")
        for module in self.__get_module_with_handler("on_privmessage"):
            module.on_privmessage(data)
        
    def on_usernotice(self, data: UserNotice):
        self.log.debug("`on_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_usernotice"):
            module.on_usernotice(data)

    def on_ritual_usernotice(self, data: RitualUserNotice):
        self.log.debug("`on_ritual_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_ritual_usernotice"):
            module.on_ritual_usernotice(data)

    def on_bitbadgeupgrade_usernotice(self, data: BitBadgeUpgradeUserNotice):
        self.log.debug("`on_bitbadgeupgrade_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_bitbadgeupgrade_usernotice"):
            module.on_bitbadgeupgrade_usernotice(data)

    def on_raid_usernotice(self, data: RaidUserNotice):
        self.log.debug("`on_raid_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_raid_usernotice"):
            module.on_raid_usernotice(data)

    def on_subscriber_usernotice(self, data: SubscriberUserNotice):
        self.log.debug("`on_subscriber_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_subscriber_usernotice"):
            module.on_subscriber_usernotice(data)
    
    def on_giftedsubscriber_usernotice(self, data: GiftedSubscriberUserNotice):
        self.log.debug("`on_giftedsubscriber_usernotice` invoked.")
        for module in self.__get_module_with_handler("on_giftedsubscriber_usernotice"):
            module.on_giftedsubscriber_usernotice(data)

    def on_command(self, data: CommandMessage):
        self.log.debug("`on_module` invoked.")
        for module in self.__get_module_with_handler("on_command"):
            module.on_command(data)






    # backend modules for managing the loading and processing of module modules
    
    def __cache_modules(self):
        modules = []
        for _file in listdir(path.join(path.dirname(__file__), 'modules/')):
            if "__init__" not in _file and "__pycache__" not in _file:
                # print ("r1: Found: ", _file)
                filename, ext = path.splitext(_file)
                if '.py' in ext:
                    modules.append(f'jarviscore.modules.{filename}')
        
        if path.exists("modules/"):
            print("Loading custom modules")
            for _file in listdir('modules/'):
                if "__init__" not in _file and "__pycache__" not in _file:
                    print ("Found: ", _file)
                    filename, ext = path.splitext(_file)
                    if '.py' in ext:
                        modules.append(f'modules.{filename}')
        
        if path.exists("bots/modules/"):
            print("Loading custom modules")
            for _file in listdir('bots/modules/'):
                if "__init__" not in _file and "__pycache__" not in _file:
                    print ("Found: ", _file)
                    filename, ext = path.splitext(_file)
                    if '.py' in ext:
                        modules.append(f'modules.{filename}')

        for extension in reversed(modules):
            try:
                # print("attempting to load: ", extension)
                self._load_module(f'{extension}')
            except Exception as e:
                raise e
                try:
                    extension = extension.replace("jarviscore", "JarvisCore")
                    print("re-attempting to load: ", extension)
                    self._load_module(f'{extension}')
                except Exception as e:
                    exc = f'{type(e).__name__}: {e}'
                    print(f'Failed to load extension {extension}\n{exc}')
        

    def _load_module(self, name):
        if name in self.__modules:
            raise ExtensionAlreadyLoaded(name)

        spec = util.find_spec(name)
        if spec is None:
            raise ExtensionNotFound(name)

        self._load_from_module_spec(spec, name)

    
    def _load_from_module_spec(self, spec, key):
        lib = util.module_from_spec(spec)
        sys.modules[key] = lib
        try:
            spec.loader.exec_module(lib)
        except Exception as e:
            del sys.modules[key]
            raise ExtensionFailed(key, e) from e

        try:
            setup = getattr(lib, 'setup')
        except AttributeError:
            del sys.modules[key]
            raise NoEntryPointError(key)

        try:
            setup(self)
        except Exception as e:
            del sys.modules[key]
            self._call_module_finalizers(lib, key)
            raise ExtensionFailed(key, e) from e
        else:
            self.__modules[key] = lib

    

    def _call_module_finalizers(self, lib, key):
        try:
            teardown = getattr(lib, 'teardown')
        except AttributeError:
            pass
        else:
            try:
                teardown(self)
            except Exception:
                pass
        finally:
            self.__modules.pop(key, None)
            sys.modules.pop(key, None)

    def __get_module_with_handler(self, handler: str):
        for module in self.__modules:
            try:
                if hasattr(self.__modules[module], handler):
                    yield self.__modules[module]
            except AttributeError:
                pass
            


            


        
        

    
